"""
utility script for optimization techniques
"""

from hyperopt import hp, fmin, tpe, Trials

# local
from foresee.scripts import fitter


def tune(train_fact, test_fact, model, params=None, args=None):
    """[summary]

    Parameters
    ----------
    train_fact : [type]
        [description]
    test_fact : [type]
        [description]
    model : [type]
        [description]

    Returns
    -------
    [type]
        [description]
    """    

    ts_train = train_fact['y']
    ts_test = test_fact['y']
    
    tune_class_obj = fitter.tuner(model)
    best, err = tune_class_obj.tune(ts_train, ts_test, params, args)
    
    return best, err


"""
Exponentially Weighted Mean
"""

import numpy as np
import pandas as pd
from hyperopt import hp, fmin, tpe, Trials

# local module
from foresee.models import models_util
from foresee.models import param_optimizer
from foresee.scripts import fitter


def ewm_fit_forecast(ts, fcst_len, params=None, args=None):
    """[summary]

    Parameters
    ----------
    ts : [type]
        [description]
    fcst_len : [type]
        [description]
    span : [type]
        [description]

    Returns
    -------
    [type]
        [description]
    """    

    try:
        alpha = params['alpha']
        
        ewm_model = ts.ewm(alpha=alpha)
        ewm_fittedvalues = ewm_model.mean()
        
        ewm_forecast = np.full(fcst_len, ewm_fittedvalues.iloc[-1])
        ewm_forecast = pd.Series(ewm_forecast)
        err = None

    except Exception as e:
        
        ewm_fittedvalues = None
        ewm_forecast = None
        err = str(e)
        
    return ewm_fittedvalues, ewm_forecast, err


def ewm_tune(ts_train, ts_test, params=None, args=None):
    
    model = 'ewm_model'
    
    try:
        
        space = hp.uniform('alpha', 0, 1)
        
        f = fitter.model_loss(model)
        
        f_obj = lambda params: f.fit_loss(ts_train, ts_test, params=None, args=None)
        
        trials = Trials()
        
        best = fmin(f_obj, space, algo=tpe.suggest, trials=trials, max_evals=100, show_progressbar=False, verbose=False)
        
        err = None
        
    except Exception as e:
        err = str(e)
        best = None
        
    return best, err
    


def ewm_main(data_dict, param_config, model_params):
    """[summary]

    Parameters
    ----------
    data_dict : [type]
        [description]
    freq : [type]
        [description]
    fcst_len : [type]
        [description]
    model_params : [type]
        [description]
    run_type : [type]
        [description]
    tune : [type]
        [description]
    epsilon : [type]
        [description]

    Returns
    -------
    [type]
        [description]
    """    

    model = 'ewm_model'
    fcst_len = param_config['FORECAST_LEN']
    output_format = param_config['OUTPUT_FORMAT']
    tune = param_config['TUNE']
    epsilon = param_config['EPSILON']
    
    complete_fact = data_dict['complete_fact']
    
    # dataframe to hold fitted values
    fitted_fact = pd.DataFrame()
    fitted_fact['y'] = complete_fact['y']
    fitted_fact['data_split'] = complete_fact['data_split']
    
    # dataframe to hold forecast values
    forecast_fact = pd.DataFrame()
    forecast_fact['y'] = np.full(fcst_len, 0)
    forecast_fact['data_split'] = np.full(fcst_len, 'Forecast')
    
    fit_fcst_fact = pd.concat([fitted_fact, forecast_fact], ignore_index=True)
    
    args = dict()
    
    # no model competition
    if output_format in ['all_models']:
        
        params = {'alpha': 0.5}
        ewm_fitted_values, ewm_forecast, err = ewm_fit_forecast(
                                                                    ts = complete_fact['y'],
                                                                    fcst_len = fcst_len,
                                                                    params = params,
                                                                )
        
        if err is None:
            fit_fcst_fact['ewm_model_forecast'] = ewm_fitted_values.append(ewm_forecast).values
            
        else:
            fit_fcst_fact['ewm_model_forecast'] = 0
            
        args['err'] = err
        args['alpha'] = params['alpha']
            
            
    # with model completition            
    if output_format in ['best_model', 'all_best']:
        
        train_fact = data_dict['train_fact']
        test_fact = data_dict['test_fact']
        
        if tune:
            # TODO: add logic when optimization fails
            opt_params, err = param_optimizer.tune(train_fact, test_fact, model)
            args['err'] = err
            
        else:
            opt_params = {'alpha': 0.5}
            
        training_fitted_values, training_forecast, training_err = ewm_fit_forecast(
                                                                                    ts = train_fact['y'],
                                                                                    fcst_len = len(test_fact) ,
                                                                                    params = opt_params
                                                                                )
        
        complete_fitted_values, complete_forecast, complete_err = ewm_fit_forecast(
                                                                                    ts = complete_fact['y'],
                                                                                    fcst_len = fcst_len,
                                                                                    params = opt_params
                                                                                )
        
        if training_err is None and complete_err is None:
            ewm_wfa = models_util.compute_wfa(
                                                y = test_fact['y'].values,
                                                yhat = training_forecast.values,
                                                epsilon = epsilon,
                                            )
            ewm_fit_fcst = training_fitted_values.append(training_forecast, ignore_index=True).append(complete_forecast, ignore_index=True)
            
            fit_fcst_fact['ewm_model_forecast'] = ewm_fit_fcst.values
            fit_fcst_fact['ewm_model_wfa'] = ewm_wfa
            
        else:
            ewm_wfa = -1
            fit_fcst_fact['ewm_model_forecast'] = 0
            fit_fcst_fact['ewm_model_wfa'] = ewm_wfa
            
        args['err'] = (training_err, complete_err)
        args['wfa'] = ewm_wfa
        
        try:
            args['alpha'] = opt_params['alpha']
        except:
            args['alpha'] = None
            
            
    return fit_fcst_fact, args



# -*- coding: utf-8 -*-
"""
Joins user input with parameters configuration and stat models configuration
and combines forecast results
"""

import os
import pandas as pd
import numpy as np
import datetime

from foresee.scripts import utils
from foresee.scripts import compose
from foresee.params import constants

# default model params
model_params = utils.read_json('model_params.json')

epsilon = constants.Constants().epsilon



def prepare_fit_report(
                        ts_df,
                        endog_colname = None,
                        ds_colname = None, 
                        gbkey = None,
                        freq = 1, 
                        fcst_length = 10, 
                        output_format = 'all_models', 
                        holdout_length = 5,
                        model_list = None,
                        processing_method = 'non_parallel',
                        tune = False,
                    ):
    """[summary]

    Parameters
    ----------
    raw_fact : [type]
        [description]
    endog_colname : [type]
        [description]
    gbkey : [type]
        [description]
    ds_column : [type]
        [description]
    freq : [type]
        [description]
    fcst_length : [type]
        [description]
    run_type : [type]
        [description]
    holdout_length : [type]
        [description]
    model_list : [type]
        [description]
    processing_method : [type]
        [description]
    tune : [type]
        [description]

    Returns
    -------
    [type]
        [description]
    """
    
    param_config = dict()
    
    param_config['Y'] = endog_colname
    param_config['DATE_STAMP'] = ds_colname
    param_config['GBKEY'] = gbkey
    param_config['FREQ'] = freq
    param_config['FORECAST_LEN'] = fcst_length
    param_config['HOLDOUT_LEN'] = holdout_length
    param_config['MODEL_LIST'] = model_list
    param_config['OUTPUT_FORMAT'] = output_format
    param_config['MODEL_LIST'] = model_list
    param_config['TUNE'] = tune
    param_config['PROCESSING_METHOD'] = processing_method
    param_config['EPSILON'] = epsilon
    
    raw_fact = ts_df.copy()
    
    pre_processed_dict = _pre_process_user_inputs(
                                                            raw_fact,
                                                            param_config,
                                                    )
    
    result, fit_result_list = compose.compose_fit(
                                                        pre_processed_dict,
                                                        model_params,
                                                        param_config,
                                                )
    
    return result, fit_result_list


def _pre_process_user_inputs(
                                    raw_fact,
                                    param_config,
                            ):
    
    endog_colname = param_config['Y']
    ds_colname = param_config['DATE_STAMP']
    output_format = param_config['OUTPUT_FORMAT']
    gbkey = param_config['GBKEY']
    holdout_length = param_config['HOLDOUT_LEN']
        
    # replace endog column name with 'y'
    
    if len(raw_fact.columns) == 1:
        raw_fact.columns = ['y']
    else:
        if endog_colname in raw_fact.columns:
            raw_fact.rename(columns={endog_colname: 'y'}, inplace=True)
        else:
            #TODO: display this error to user
            raise ValueError('time series column name not found!!!')

    
    # add or rename date-time column
    
    if ds_colname not in raw_fact.columns:
        raw_fact['date_stamp'] = pd.date_range(end=datetime.datetime.now(), periods=len(raw_fact), freq='D')
        
    else:
        raw_fact.rename(columns={ds_colname: 'date_stamp'}, inplace=True)
        raw_fact['date_stamp'] = pd.to_datetime(raw_fact['date_stamp'])
#        raw_fact['date_stamp'] = raw_fact['date_stamp'].astype('datetime64[ns]')
        
    #TODO: if user chose to compare models, then create train-holdout sets
    #TODO: missing data interpolation, needs user input!
    
    pre_processed_dict = dict()    
        
    if output_format in ['best_model', 'all_best']:
        
        if gbkey in raw_fact.columns:
            for k,v in raw_fact.groupby(gbkey):
                train_fact = v.iloc[:-holdout_length]
                test_fact = v.iloc[-holdout_length:]
                v['data_split'] = np.concatenate([np.full(len(train_fact), 'Train'), np.full(len(test_fact), 'Test')])

                pre_processed_dict[k] = {'complete_fact': v, 'train_fact': train_fact, 'test_fact': test_fact}
                
        else:
            train_fact = raw_fact.iloc[:-holdout_length]
            test_fact = raw_fact.iloc[-holdout_length:]
            raw_fact['data_split'] = np.concatenate([np.full(len(train_fact), 'Train'), np.full(len(test_fact), 'Test')])
            
            pre_processed_dict[1] = {'complete_fact': raw_fact, 'train_fact': train_fact, 'test_fact': test_fact}
            
            
    else:
        if gbkey in raw_fact.columns:
            for k,v in raw_fact.groupby(gbkey):
                v['data_split'] = 'Train'
                pre_processed_dict[k] = {'complete_fact': v}
                
        else:
            raw_fact['data_split'] = 'Train'
            pre_processed_dict[1] = {'complete_fact': raw_fact}
            
    
    return pre_processed_dict
  






"""
Holt-Winters from statsmodels
"""

import warnings
warnings.filterwarnings('ignore')

import numpy as np
import pandas as pd
from statsmodels.tsa.holtwinters import ExponentialSmoothing
from hyperopt import hp, fmin, tpe, Trials

# local module
from foresee.models import models_util
from foresee.models import param_optimizer
from foresee.scripts import fitter


def holt_winters_fit_forecast(ts, fcst_len, params=None, args=None):
    """[summary]

    Parameters
    ----------
    ts : [type]
        [description]
    fcst_len : [type]
        [description]
    freq : [type]
        [description]
    params : [type]
        [description]

    Returns
    -------
    [type]
        [description]
    """
    
    freq = args['FREQ']
    
    try:
        if params is None:
            hw_model = ExponentialSmoothing(
                                                endog = ts,
                                                seasonal_periods = freq
                                              ).fit(optimized=True)
        else:
            hw_model = ExponentialSmoothing(
                                                endog = ts,
                                                trend = params['trend'],
                                                seasonal = params['seasonal'],
                                                seasonal_periods = freq
                                              ).fit(optimized=True)
            
        hw_fittedvalues = hw_model.fittedvalues    

        hw_forecast = hw_model.predict(
                                            start = len(ts),
                                            end =   len(ts) + fcst_len - 1
                                        )
        err = None
        
    except Exception as e:
        hw_fittedvalues = None
        hw_forecast = None
        err = str(e)
    
    
    return hw_fittedvalues, hw_forecast, err


def holt_winters_tune(ts_train, ts_test, params=None, args=None):
    
    model = 'holt_winters'
    
    try:
        options = ['add', 'mul']
        space = {
                    'trend': hp.choice('trend', options),
                    'seasonal': hp.choice('seasonal', options)
                }
        
        f = fitter.model_loss(model)
        f_obj = lambda params: f.fit_loss(ts_train, ts_test, params, args)
        
        trials = Trials()
        best_raw = fmin(f_obj, space, algo=tpe.suggest, trials=trials, max_evals=10, show_progressbar=False, verbose=False)
        best = {'trend': options[best_raw['trend']], 'seasonal': options[best_raw['seasonal']]}
        err = None
        
    except Exception as e:
        err = str(e)
        best = None
        
        
    return best, err


def holt_winters_main(data_dict, param_config, model_params):
    """[summary]

    Parameters
    ----------
    data_dict : [type]
        [description]
    freq : [type]
        [description]
    fcst_len : [type]
        [description]
    model_params : [type]
        [description]
    run_type : [type]
        [description]
    tune : [type]
        [description]
    epsilon : [type]
        [description]

    Returns
    -------
    [type]
        [description]
    """    

    model = 'holt_winters'
    fcst_len = param_config['FORECAST_LEN']
    output_format = param_config['OUTPUT_FORMAT']
    tune = param_config['TUNE']
    epsilon = param_config['EPSILON']  
    freq = param_config['FREQ']
    
    complete_fact = data_dict['complete_fact']
    
    # dataframe to hold fitted values
    fitted_fact = pd.DataFrame()
    fitted_fact['y'] = complete_fact['y']
    fitted_fact['data_split'] = complete_fact['data_split']
    
    # dataframe to hold forecast values
    forecast_fact = pd.DataFrame()
    forecast_fact['y'] = np.full(fcst_len, 0)
    forecast_fact['data_split'] = np.full(fcst_len, 'Forecast')
    
    fit_fcst_fact = pd.concat([fitted_fact, forecast_fact], ignore_index=True)
    
    args = dict()
    args['FREQ'] =  freq
    
    fit_args = dict()

    # no model competition    
    if output_format in ['all_models']:
        
        fitted_values, forecast, err = holt_winters_fit_forecast(
                                                                       ts = complete_fact['y'],
                                                                       fcst_len = fcst_len,
                                                                       params = None,
                                                                       args = args
                                                                  )
        if err is None:
            fit_fcst_fact['holt_winters_forecast'] = fitted_values.append(forecast).values
            
        else:
            fit_fcst_fact['holt_winters_forecast'] = 0
            
        fit_args['err'] = err
    
    
    # with model completition            
    if output_format in ['best_model', 'all_best']:
        
        train_fact = data_dict['train_fact']
        test_fact = data_dict['test_fact']
        
        if tune:
            # TODO: add logic when optimization fails
            
            params, err = param_optimizer.tune(train_fact, test_fact, model, params=None, args=args)
            
            fit_args['tune_err'] = err
            
        else:
            params = None
            
        training_fitted_values, holdout_forecast, training_err = holt_winters_fit_forecast(
                                                                                                   ts = train_fact['y'],
                                                                                                   fcst_len = len(test_fact),
                                                                                                   params = params,
                                                                                                   args = args,
                                                                                         )
        
        complete_fitted_values, complete_forecast, complete_err = holt_winters_fit_forecast(
                                                                                                   ts = complete_fact['y'],
                                                                                                   fcst_len = fcst_len,
                                                                                                   params = params,
                                                                                                   args = args,
                                                                                         )
        
        if training_err is None and complete_err is None:
            hw_wfa = models_util.compute_wfa(
                                                y = test_fact['y'].values,
                                                yhat = holdout_forecast.values,
                                                epsilon = epsilon,
                                            )
            hw_fit_fcst = training_fitted_values.append(holdout_forecast, ignore_index=True).append(complete_forecast, ignore_index=True)
            
            fit_fcst_fact['holt_winters_forecast'] = hw_fit_fcst.values
            fit_fcst_fact['holt_winters_wfa'] = hw_wfa
            
        else:
            hw_wfa = -1
            fit_fcst_fact['holt_winters_forecast'] = 0
            fit_fcst_fact['holt_winters_wfa'] = -1
            
        fit_args['err'] = (training_err, complete_err)
        fit_args['wfa'] = hw_wfa
        fit_args['params'] = params
        
        
    return fit_fcst_fact, fit_args







# -*- coding: utf-8 -*-
"""
fitter class
"""

import os
import sys
from sklearn.metrics import mean_absolute_error, mean_squared_error

# import local modules

from foresee.models.ewm import ewm_main, ewm_fit_forecast, ewm_tune
from foresee.models.fft import fft_main, fft_fit_forecast, fft_tune
from foresee.models.holt_winters import holt_winters_main, holt_winters_fit_forecast, holt_winters_tune
from foresee.models.prophet import prophet_main, prophet_fit_forecast, prophet_tune
from foresee.models.sarimax import sarimax_main, sarimax_fit_forecast, sarimax_tune

class fitter:
    """[summary]

    Returns
    -------
    [type]
        [description]
    """    
    FIT_MODELS = {
                    'ewm_model':        ewm_main,
                    'fft':              fft_main,
                    'holt_winters':     holt_winters_main,
                    'prophet':          prophet_main,
                    'sarimax':          sarimax_main,
                 }
    
    
    def __init__(self, model):
         self.model = model
         
    
    def fit(self, data_dict, param_config, model_params):
        """[summary]

        Parameters
        ----------
        data_dict : [type]
            [description]
        freq : [type]
            [description]
        forecast_len : [type]
            [description]
        model_params : [type]
            [description]
        run_type : [type]
            [description]
        tune : [type]
            [description]
        epsilon : [type]
            [description]

        Returns
        -------
        [type]
            [description]
        """        
        fit_model = self.FIT_MODELS[self.model]
        
        return fit_model(data_dict, param_config, model_params)
    
class tuner:
    
    TUNE_MODELS = {
                        'ewm_model':        ewm_tune,
                        'fft':              fft_tune,
                        'holt_winters':     holt_winters_tune,
                        'prophet':          prophet_tune,
                        'sarimax':          sarimax_tune,           
                    }
    
    def __init__(self, model):
        self.model = model
        
    def tune(self, ts_train, ts_test, params, args):
        
        tune_model = self.TUNE_MODELS[self.model]
        
        return tune_model(ts_train, ts_test, params, args)
    
    
    
class model_loss:
    
    FIT_MODELS = {
                        'ewm_model':        ewm_fit_forecast,
                        'fft':              fft_fit_forecast,
                        'holt_winters':     holt_winters_fit_forecast,
                        'prophet':          prophet_fit_forecast,           
                        'sarimax':          sarimax_fit_forecast,           
                    }
    
    def __init__(self, model):
        self.model = model
        
    def fit_loss(self, ts_train, ts_test, params=None, args=None):
        
        fit_model = self.FIT_MODELS[self.model]
        
        fcst_len = len(ts_test)
        
        fitted_values, yhat, err = fit_model(ts_train, fcst_len, params, args)
        
        if err is None:
            loss = mean_absolute_error(ts_test.values, yhat.values)
        else:
            loss = ts_test.abs().mean()
        
        
        return loss


    


"""
Prophet from fbprophet
"""

import warnings
warnings.filterwarnings('ignore')

import numpy as np
import pandas as pd
import os
from fbprophet import Prophet
from hyperopt import hp, fmin, tpe, Trials

# local module
from foresee.models import models_util
from foresee.models import param_optimizer
from foresee.scripts import fitter


class suppress_stdout_stderr(object):
    """[summary]

    Parameters
    ----------
    object : [type]
        [description]
    """

    def __init__(self):
        # Open a pair of null files
        self.null_fds = [os.open(os.devnull, os.O_RDWR) for x in range(2)]
        # Save the actual stdout (1) and stderr (2) file descriptors.
        self.save_fds = [os.dup(1), os.dup(2)]

    def __enter__(self):
        # Assign the null pointers to stdout and stderr.
        os.dup2(self.null_fds[0], 1)
        os.dup2(self.null_fds[1], 2)

    def __exit__(self, *_):
        # Re-assign the real stdout/stderr back to (1) and (2)
        os.dup2(self.save_fds[0], 1)
        os.dup2(self.save_fds[1], 2)
        # Close the null files
        for fd in self.null_fds + self.save_fds:
            os.close(fd)

            
def prophet_fit_forecast(df, fcst_len, params=None, args=None):
    """[summary]

    Parameters
    ----------
    df : [type]
        [description]
    fcst_len : [type]
        [description]
    freq : [type]
        [description]

    Returns
    -------
    [type]
        [description]
    """    

    df = df[['date_stamp','y']].reset_index(drop=True)
    df.columns = ['ds', 'y']
    
    freq = args['FREQ']
    
    if freq == 12:
        prophet_freq = 'M'
    elif freq == 52:
        prophet_freq = 'W'
    else:
        prophet_freq = None
    
    try:
        
        prophet_model = Prophet()
        
        with suppress_stdout_stderr():
            prophet_model.fit(df)
            
        future = prophet_model.make_future_dataframe(periods=fcst_len, freq=prophet_freq, include_history=True)
        
        prophet_model_predictions = prophet_model.predict(future)['yhat']
        
        prophet_fittedvalues = prophet_model_predictions.head(len(df))                
        
        prophet_forecast = prophet_model_predictions.tail(fcst_len)
        
        err = None
        
    except Exception as e:
        
        prophet_fittedvalues = None
        prophet_forecast = None
        err = str(e)
        
    return prophet_fittedvalues, prophet_forecast, err


def prophet_tune(ts_train, ts_test, params=None, args=None):
    return None

def prophet_main(data_dict, param_config, model_params):
    """[summary]

    Parameters
    ----------
    data_dict : [type]
        [description]
    freq : [type]
        [description]
    fcst_len : [type]
        [description]
    model_params : [type]
        [description]
    run_type : [type]
        [description]
    tune : [type]
        [description]
    epsilon : [type]
        [description]

    Returns
    -------
    [type]
        [description]
    """    

    model = 'prophet'
    fcst_len = param_config['FORECAST_LEN']
    output_format = param_config['OUTPUT_FORMAT']
    tune = param_config['TUNE']
    epsilon = param_config['EPSILON']
    freq = param_config['FREQ']
    
    complete_fact = data_dict['complete_fact']
    
    # dataframe to hold fitted values
    fitted_fact = pd.DataFrame()
    fitted_fact['y'] = complete_fact['y']
    fitted_fact['data_split'] = complete_fact['data_split']
    
    # dataframe to hold forecast values
    forecast_fact = pd.DataFrame()
    forecast_fact['y'] = np.full(fcst_len, 0)
    forecast_fact['data_split'] = np.full(fcst_len, 'Forecast')
    
    fit_fcst_fact = pd.concat([fitted_fact, forecast_fact], ignore_index=True)
    
    args = dict()
    args['FREQ'] = freq
    
    fit_args = dict()
    
    if output_format in ['all_models']:
    
        prophet_fitted_values, prophet_forecast, err = prophet_fit_forecast(
                                                                                df = complete_fact,
                                                                                fcst_len = fcst_len,
                                                                                args = args
                                                                            )
        if err is None:
            fit_fcst_fact['prophet_forecast'] = prophet_fitted_values.append(prophet_forecast).values
            
        else:
            fit_fcst_fact['prophet_forecast'] = 0
            
        fit_args['err'] = err
            
            
    if output_format in ['best_model', 'all_best']:
        
        train_fact = data_dict['train_fact']
        test_fact = data_dict['test_fact']
        
        training_fitted_values, training_forecast, training_err = prophet_fit_forecast(
                                                                                            df = train_fact,
                                                                                            fcst_len = len(test_fact),
                                                                                            args = args
                                                                                    )
        
        complete_fitted_values, complete_forecast, complete_err = prophet_fit_forecast(
                                                                                            df = complete_fact,
                                                                                            fcst_len = fcst_len,
                                                                                            args = args,
                                                                                    )
        
        if training_err is None and complete_err is None:
            prophet_wfa = models_util.compute_wfa(
                                                    y = test_fact['y'].values,
                                                    yhat = training_forecast.values,
                                                    epsilon = epsilon
                                                )
            
            prophet_fit_fcst = training_fitted_values.append(training_forecast, ignore_index=True).append(complete_forecast, ignore_index=True)
            
            fit_fcst_fact['prophet_forecast'] = prophet_fit_fcst.values
            fit_fcst_fact['prophet_wfa'] = prophet_wfa
            
        else:
            prophet_wfa = -1
            fit_fcst_fact['prophet_forecast'] = 0
            fit_fcst_fact['prophet_wfa'] = prophet_wfa
            
        fit_args['err'] = (training_err, complete_err)
        fit_args['wfa'] = prophet_wfa
            
            
    return fit_fcst_fact, fit_args




# -*- coding: utf-8 -*-
"""
Controls input and output of forecasting process
"""

import pandas as pd
import numpy as np
import datetime
import dask
dask.config.set(scheduler='processes')

from foresee.scripts import fitter

def generate_fit_forecast(dict_key, dict_values, param_config, model_params):
    """[summary]

    Parameters
    ----------
    dict_key : [type]
        [description]
    dict_values : [type]
        [description]
    model_list : [type]
        [description]
    freq : [type]
        [description]
    forecast_len : [type]
        [description]
    model_params : [type]
        [description]
    run_type : [type]
        [description]
    tune : [type]
        [description]
    epsilon : [type]
        [description]

    Returns
    -------
    [type]
        [description]
    """    
    
    model_list = param_config['MODEL_LIST']
    
    fit_result = dict()
    fit_result['ts_id'] = dict_key
    
    for m in model_list:

        f = fitter.fitter(m)

        (
         fit_result[m+'_fit_fcst_df'],
         fit_result[m+'_args']
         ) = f.fit(dict_values, param_config, model_params)
    
    return fit_result


def compose_fit(pre_processed_dict, model_params, param_config):
    """[summary]

    Parameters
    ----------
    pre_processed_dict : [type]
        [description]
    model_params : [type]
        [description]
    param_config : [type]
        [description]
    gbkey : [type]
        [description]
    run_type : [type]
        [description]
    processing_method : [type]
        [description]
    tune : [type]
        [description]

    Returns
    -------
    [type]
        [description]
    """    

#    freq = param_config['FREQ']
#    forecast_len = param_config['FORECAST_LEN']
#    model_list = param_config['MODEL_LIST']
#    epsilon = param_config['EPSILON']

    processing_method = param_config['PROCESSING_METHOD']
    model_list = param_config['MODEL_LIST']
    output_format = param_config['OUTPUT_FORMAT']
    
    # non-parallel fit
    
    if processing_method == 'parallel':
        task_list = list()
        
        for dict_key, dict_values in pre_processed_dict.items():
            fit_task = dask.delayed(generate_fit_forecast)(
                                                               dict_key,
                                                               dict_values,
                                                               param_config,
                                                               model_params,
                                                          )
            task_list.append(fit_task)

        fit_result_list = dask.compute(task_list)[0]
        
    else:
        fit_result_list = list()

        for dict_key, dict_values  in pre_processed_dict.items():
            fit_result = generate_fit_forecast(
                                                    dict_key,
                                                    dict_values,
                                                    param_config,
                                                    model_params,
                                            )
            fit_result_list.append(fit_result)
        
    result = combine_to_dataframe(fit_result_list, model_list, output_format)
        
    return result, fit_result_list


def combine_to_dataframe(fit_result_list, model_list, output_format):
    """[summary]

    Parameters
    ----------
    fit_result_list : [type]
        [description]
    model_list : [type]
        [description]
    run_type : [type]
        [description]

    Returns
    -------
    [type]
        [description]
    """

    if output_format == 'best_model':

        fit_result_list = _find_best_model(fit_result_list, model_list)
        
        df_list = list()

        for result_dict in fit_result_list:

            bm = result_dict['best_model']
            df = result_dict[bm+'_fit_fcst_df']
            
            df['best_model'] = bm
            df['ts_id'] = result_dict['ts_id']
            df_list.append(df)

        result = pd.concat(df_list, axis=0, ignore_index=True)
        
    elif output_format == 'all_best':
        
        fit_result_list = _find_best_model(fit_result_list, model_list)
        
        df_list = list()

        for result_dict in fit_result_list:
            bm = result_dict['best_model']
            df = result_dict[bm+'_fit_fcst_df']
            
            for m in model_list:
                mdf = result_dict[m+'_fit_fcst_df']
                df[m+'_forecast'] = mdf[m+'_forecast'].values
                df[m+'_wfa'] = mdf[m+'_wfa'].values

            df['ts_id'] = result_dict['ts_id']
            df['best_model'] = bm
            df_list.append(df)

        result = pd.concat(df_list, axis=0, ignore_index=True)
                
        
    else:
        df_list = list()
        
        for result_dict in fit_result_list:
            m1 = model_list[0]
            df = result_dict[m1+'_fit_fcst_df']
            
            for m in model_list[1:]:
                mdf = result_dict[m+'_fit_fcst_df']
                df[m+'_forecast'] = mdf[m+'_forecast'].values
                
            df['ts_id'] = result_dict['ts_id']
            df_list.append(df)
        
        result = pd.concat(df_list, axis=0, ignore_index=True) 
    
    return result


def _find_best_model(fit_result_list, model_list):
    """[summary]

    Parameters
    ----------
    fit_result_list : [type]
        [description]
    model_list : [type]
        [description]

    Returns
    -------
    [type]
        [description]
    """    

    for result_dict in fit_result_list:
        model_wfa_dict = dict()
        
        for m in model_list:
            model_wfa_dict[m] = result_dict[m+'_args']['wfa']
            
        result_dict['best_model'] = [k for k,v in model_wfa_dict.items() if v == max(model_wfa_dict.values())][0]
    
    return fit_result_list


# not in use
def _transform_dataframe_to_dict(raw_fact, gbkey):
    """[summary]

    Parameters
    ----------
    raw_fact : [type]
        [description]
    gbkey : [type]
        [description]

    Returns
    -------
    [type]
        [description]
    """    

    data_param_list = list()
    
    if gbkey is None:
        
        data_param_list.append({'ts_id':1, 'df':raw_fact})
    
    else:
        for k,v in raw_fact.groupby(gbkey):
            data_param_list.append({'ts_id':k, 'df':v})
    
    return data_param_list


